// -*- C++ -*-
/*!
 * @file  UVCCamera.cpp
 * @brief USB Video Class(UVC) cameras module Acquire component
 * @date $Date$
 *
 * @author メディアドライブ
 * rtc-dev@mediadrive.jp
 *
 * BSD 3-clause "New" or "Revised" License
 *
 * $Id$
 */

#include "UVCCamera.h"
#include "ImageDataConvert.h"

cv::VideoCapture cap;

// Module specification
// <rtc-template block="module_spec">
static const char* uvccamera_spec[] =
  {
    "implementation_id", "UVCCamera",
    "type_name",         "UVCCamera",
    "description",       "USB Video Class(UVC) cameras module Acquire component",
    "version",           "1.0.0",
    "vendor",            "MEDIA DRIVE",
    "category",          "ImageProcessing",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.upsidedownMode", "off",
    "conf.default.width", "640",
    "conf.default.height", "480",
    "conf.default.compress", "jpeg",
    "conf.default.quality", "95",
    "conf.default.cameraID", "0",

    // Widget
    "conf.__widget__.upsidedownMode", "radio",
    "conf.__widget__.width", "spin",
    "conf.__widget__.height", "spin",
    "conf.__widget__.compress", "radio",
    "conf.__widget__.quality", "spin",
    "conf.__widget__.cameraID", "text",
    // Constraints
    "conf.__constraints__.upsidedownMode", "(off,on)",
    "conf.__constraints__.compress", "(off,jpeg)",
    "conf.__constraints__.quality", "0<=x<=100",

    "conf.__type__.upsidedownMode", "string",
    "conf.__type__.width", "int",
    "conf.__type__.height", "int",
    "conf.__type__.compress", "string",
    "conf.__type__.quality", "int",
    "conf.__type__.cameraID", "int",

    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
UVCCamera::UVCCamera(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_CaptureImageOutOut("CaptureImageOut", m_CaptureImageOut)

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
UVCCamera::~UVCCamera()
{
}


/*!
 * 各変数の初期化を行います
 */
RTC::ReturnCode_t UVCCamera::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  
  // Set OutPort buffer
  addOutPort("CaptureImageOut", m_CaptureImageOutOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("upsidedownMode", m_upsidedownMode, "off");
  bindParameter("width", m_width, "640");
  bindParameter("height", m_height, "480");
  bindParameter("compress", m_compress, "jpeg");
  bindParameter("quality", m_quality, "95");
  bindParameter("cameraID", m_cameraID, "0");
  // </rtc-template>
  
  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t UVCCamera::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t UVCCamera::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t UVCCamera::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*!
 * カメラの接続準備を行います
 */

RTC::ReturnCode_t UVCCamera::onActivated(RTC::UniqueId ec_id)
{
  cap = cv::VideoCapture (m_cameraID);
    if(!cap.isOpened())//カメラデバイスが正常にオープンしたか確認．
    {
        //読み込みに失敗したときの処理
        return RTC::RTC_ERROR;
    }
	if (!cap.set(CV_CAP_PROP_FRAME_WIDTH, m_width))
	{
		return RTC::RTC_ERROR;
	}

	if(!cap.set(CV_CAP_PROP_FRAME_HEIGHT, m_height))
	{
		return RTC::RTC_ERROR;
	}
	return RTC::RTC_OK;
}

/*!
 * カメラ用メモリの開放を行います
 */

RTC::ReturnCode_t UVCCamera::onDeactivated(RTC::UniqueId ec_id)
{
  //カメラ用メモリの解放
  if(cap.isOpened())
  {
        cap.release();
  }
  return RTC::RTC_OK;
}

/*!
 * カメラで撮影した画像をOutPortに送ります
 */

RTC::ReturnCode_t UVCCamera::onExecute(RTC::UniqueId ec_id)
{
    static coil::TimeValue tm_pre;
    static int count = 0;
    cv::Mat input_image;

    if(cap.isOpened() && cap.get(CV_CAP_PROP_FRAME_WIDTH)  != -1 && cap.get(CV_CAP_PROP_FRAME_HEIGHT) != -1) 
    {
        cv::Mat cam_raw;
        cap >> cam_raw;
          
        cv::Mat cam_edit;

        if(m_upsidedownMode != "off")
        {
          cv::flip(cam_raw, cam_edit, 0);
        }
        else
        {
         cam_edit = cam_raw;
        }

        SetCameraImage(m_CaptureImageOut, cam_edit, m_compress, m_quality);


        m_CaptureImageOutOut.write();

            if (count > 100)
            {
                count = 0;
                coil::TimeValue tm;
                tm = coil::gettimeofday();

                double sec(tm - tm_pre);

                if (sec > 1.0 && sec < 1000.0)
                {
                        std::cout << 100/sec << " [FPS]" << std::endl;
                }

                tm_pre = tm;
            }
            ++count;
        }

        return RTC::RTC_OK;

}

/*
RTC::ReturnCode_t UVCCamera::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t UVCCamera::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t UVCCamera::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t UVCCamera::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t UVCCamera::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/



extern "C"
{
 
  void UVCCameraInit(RTC::Manager* manager)
  {
    coil::Properties profile(uvccamera_spec);
    manager->registerFactory(profile,
                             RTC::Create<UVCCamera>,
                             RTC::Delete<UVCCamera>);
  }
  
};


